# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List, Optional
from functools import wraps
import logging

from .threshold_type import ThresholdType
from .error.failure import PromptTestSetupException
from .runner.prompt_template_runner import (
    run_prompt_template_test,
    run_prompt_template_test_from_files,
)
from .runner.system_prompt_runner import (
    run_system_prompt_test,
    run_system_prompt_test_from_files,
)

TESTS_TO_RUN = []


def prompttest(
    experiment_classname: Callable,
    model_name: str,
    metric_name: str,
    threshold: float = 1.0,
    prompt_template_file: str = None,
    user_input_file: str = None,
    system_prompt_file: str = None,
    human_messages_file: str = None,
    prompt_template: str = None,
    user_input: Optional[List[Dict[str, str]]] = None,
    system_prompt: str = None,
    human_messages: Optional[List[str]] = None,
    is_average: bool = False,
    use_input_pairs: bool = True,
    threshold_type: ThresholdType = ThresholdType.MAXIMUM,
    model_arguments: Dict[str, object] = {},
):
    r"""
    Creates a decorator for prompt tests, which can annotate evaluation functions.
    This enables developers to create a prompt test suite from their evaluations.
    """

    def prompttest_decorator(eval_fn: Callable):
        @wraps(eval_fn)
        def runs_test():
            if prompt_template_file and user_input_file:
                return run_prompt_template_test_from_files(
                    experiment_classname,
                    model_name,
                    metric_name,
                    eval_fn,
                    threshold,
                    threshold_type,
                    is_average,
                    prompt_template_file,
                    user_input_file,
                    use_input_pairs,
                    model_arguments,
                )
            elif prompt_template and user_input:
                return run_prompt_template_test(
                    experiment_classname,
                    model_name,
                    metric_name,
                    eval_fn,
                    threshold,
                    threshold_type,
                    is_average,
                    prompt_template,
                    user_input,
                    use_input_pairs,
                    model_arguments,
                )
            elif system_prompt_file and human_messages_file:
                return run_system_prompt_test_from_files(
                    experiment_classname,
                    model_name,
                    metric_name,
                    eval_fn,
                    threshold,
                    threshold_type,
                    is_average,
                    system_prompt_file,
                    human_messages_file,
                    use_input_pairs,
                    model_arguments,
                )
            elif system_prompt and human_messages:
                return run_system_prompt_test(
                    experiment_classname,
                    model_name,
                    metric_name,
                    eval_fn,
                    threshold,
                    threshold_type,
                    is_average,
                    system_prompt,
                    human_messages,
                    use_input_pairs,
                    model_arguments,
                )
            else:
                logging.error("Bad configuration for metric: " + metric_name)
                raise PromptTestSetupException

        TESTS_TO_RUN.append(runs_test)
        return runs_test

    return prompttest_decorator


def main():
    logging.getLogger().setLevel(logging.WARNING)
    print("Running " + str(len(TESTS_TO_RUN)) + " test(s)")
    failures = sum([test() for test in TESTS_TO_RUN])
    if failures == 0:
        print("All " + str(len(TESTS_TO_RUN)) + " test(s) passed!")
        exit(0)
    else:
        print("Tests failed: " + str(failures))
        exit(1)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

r"""
These mock functions exist for testing and demo purposes.
"""


def mock_chat_completion_fn(**kwargs):
    return {
        "choices": [
            {
                "finish_reason": "stop",
                "index": 0,
                "message": {
                    "content": "George Washington",
                    "role": "assistant",
                },
            }
        ],
        "created": 1687839008,
        "id": "",
        "model": "gpt-3.5-turbo-0301",
        "object": "chat.completion",
        "usage": {"completion_tokens": 18, "prompt_tokens": 57, "total_tokens": 75},
    }


def mock_completion_fn(**kwargs):
    return {
        "id": "",
        "object": "text_completion",
        "created": 1589478378,
        "model": "text-davinci-003",
        "choices": [
            {
                "text": "\n\nGeorge Washington",
                "index": 0,
                "logprobs": None,
                "finish_reason": "length",
            }
        ],
        "usage": {"prompt_tokens": 5, "completion_tokens": 7, "total_tokens": 12},
    }


def mock_hf_completion_fn(**kwargs):
    return {"generated_text": "George Washington"}


# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from typing import Callable, Dict, List, Tuple


class PromptTestRunner:
    r"""
    Base class for prompt test runners. Please use the subclass instead.s
    """

    def __init__(self):
        self.ran = defaultdict(bool)
        self.harnesses = dict()

    def run(self, *args, **kwargs) -> str:
        r"""
        Runs the test if it has not already been run.
        """
        key = str(args)
        if self.ran[key]:
            return key
        self.harnesses[key] = self._get_harness(*args, **kwargs)
        self.harnesses[key].run()
        self.ran[key] = True
        return key

    def evaluate(
        self,
        key: str,
        metric_name: str,
        eval_fn: Callable,
        use_input_pairs: Dict[str, Tuple[str, Dict[str, str]]],
    ) -> None:
        r"""
        Evaluates the test results.
        """
        self.harnesses[key].evaluate(metric_name, eval_fn, use_input_pairs)

    def rank(self, key: str, metric_name: str, is_average: bool) -> Dict[str, float]:
        r"""
        Uses evaluations to "rank" the template or system prompt.
        This creates an overall score (sum or average) across all
        test inputs, which can be compared against a threshold.
        """
        return self.harnesses[key].rank(metric_name, is_average)

    @staticmethod
    def _get_harness(
        experiment_classname,
        model_name: str,
        prompt_template: str,
        user_inputs: List[Dict[str, str]],
        model_args: Dict[str, object],
    ):
        raise NotImplementedError(
            "This should be implemented by a subclass of `PromptTestRunner`."
        )



import os
from .error import PromptToolsUtilityError
from pylint import epylint as lint
import logging

PROMPTTOOLS_TMP = "prompttools_tmp.py"

def validate(text: str):
    r"""
    Validates that the generated text is python.

    Args:
        text (str): The generated text, which should be valid python.
    """
    if os.path.isfile(PROMPTTOOLS_TMP):
        raise PromptToolsUtilityError
    with open(PROMPTTOOLS_TMP, "w") as f:
        f.write(text)
    pylint_stdout, _ = lint.py_run(PROMPTTOOLS_TMP, return_std=True)
    os.remove(PROMPTTOOLS_TMP)
    return 0.0 if "error" in pylint_stdout.getvalue() else 1.0







# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, List
import pandas as pd
from IPython import display
import ipywidgets as widgets


class ComparisonWidgetProvider:
    r"""
    Provides functionality for widgets to compare models. This includes
    displaying widgets, and recording evaluations in the experiment.
    """

    def __init__(self, completion_fn, agg_fn, eval_listener_fn):
        self.completion_fn = completion_fn
        self.agg_fn = agg_fn
        self.eval_listener_fn = eval_listener_fn

    def _get_comparison_submission_listener(
        self, table: pd.DataFrame, models: List[str]
    ) -> Callable:
        def on_click(b):
            sorted_scores = self.agg_fn(table, 0)
            data = {
                models[0]: sorted_scores.keys(),
                "feedback": sorted_scores.values(),
            }
            df = pd.DataFrame(data)
            display.display(df)

        return on_click

    def set_models(self, models: List[str]) -> None:
        self.models = models
        self.row_len = 2 + len(self.models)

    def get_header_widgets(self) -> List[object]:
        return (
            [widgets.Label("Input")]
            + [widgets.Label(model) for model in self.models]
            + [widgets.Label("Feedback")]
        )

    def get_row_widgets(self, index, row):
        items = [
            widgets.HTML(
                value="<style>p{word-wrap: break-word}</style> <p>" + row.name + " </p>"
            )
        ]
        items += [
            widgets.HTML(
                value="<style>p{word-wrap: break-word}</style> <p>"
                + row[model][0]
                + " </p>"
            )
            for model in self.models
        ]
        feedback_dropdown = widgets.Dropdown(
            options=[("\U0001F44D", 1), ("\U0001F44E", 0)],
            value=1,
            layout={"width": "50px"},
        )
        feedback_dropdown.observe(self.eval_listener_fn(index), names="value")
        items += [feedback_dropdown]
        return items

    def get_footer_widgets(self, table):
        submit_button = widgets.Button(
            description="Submit",
            disabled=False,
            button_style="success",
            tooltip="Submit",
        )
        submit_button.on_click(
            self._get_comparison_submission_listener(table, self.models)
        )
        return [widgets.Label("")] * (self.row_len - 1) + [submit_button]

    def display(self, items):
        row_len = 2 + len(self.models)
        grid = widgets.GridBox(
            items,
            layout=widgets.Layout(
                grid_template_columns="repeat(" + str(row_len) + ", 230px)"
            ),
        )
        display.display(grid)



import os
import openai
import jinja2
from .error import PromptToolsUtilityError

EVALUATION_SYSTEM_PROMPT="""
Determine whether or not the response is following directions.
Your answer should either be "RIGHT" if the response follows directions,
or "WRONG" if the model is not following directions.
"""

EVALUATION_USER_TEMPLATE="""
PROMPT: {{prompt}}
RESPONSE: {{response}}
ANSWER: 
"""

def _get_messages(prompt: str, response: str):
    environment = jinja2.Environment()
    template = environment.from_string(EVALUATION_USER_TEMPLATE)
    user_message = template.render({"prompt": prompt, 
                                       "response": response})
    return [{"role": "system", "content": EVALUATION_SYSTEM_PROMPT},
            {"role": "user", "content": user_message}]


def compute(prompt: str, response: str, model: str = 'gpt-4') -> float:
    r"""
    Uses a high quality chat model, like GPT-4, to automatically evaluate a given
    prompt/response pair. Outputs can be 0 or 1.

    Args:
        prompt (str): The input prompt.
        response (str): The model response.
        model (str): The OpenAI chat model to use for generating an expected response.
            Defaults to GPT-4.
    """
    if not os.environ['OPENAI_API_KEY']:
        raise PromptToolsUtilityError
    evaluation = openai.ChatCompletion.create(model=model,
                                              messages=_get_messages(prompt, response))
    return 1.0 if 'RIGHT' in evaluation['choices'][0]['message']['content'] else 0.0


# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


class PromptToolsUtilityError(Exception):
    r"""
    An exception to throw when something goes wrong with the prompttools utility.
    """

    pass



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List
from prompttools.experiment import Experiment


class ExperimentationHarness:
    r"""
    Base class for experimentation harnesses. This should not be used directly, please use the subclasses instead.
    """

    experiment: Experiment
    PIVOT_COLUMNS: list

    def __init__(self) -> None:
        self.input_pairs_dict = None
        self.experiment = None
        self.runs = 1

    @staticmethod
    def _prepare_arguments(arguments: Dict[str, object]) -> Dict[str, List[object]]:
        return {name: [arg] for name, arg in arguments.items()}

    def prepare(self) -> None:
        r"""
        Prepares the underlying experiment.
        """
        self.experiment.prepare()

    def run(self) -> None:
        r"""
        Runs the underlying experiment.
        """
        if self.input_pairs_dict:
            self.experiment.run(
                self.PIVOT_COLUMNS[0], self.input_pairs_dict, runs=self.runs
            )
        else:
            self.experiment.run(runs=self.runs)

    def evaluate(
        self, metric_name: str, eval_fn: Callable, use_input_pairs: bool = True
    ) -> None:
        r"""
        Uses the given eval_fn to evaluate the results of the underlying experiment.
        """
        if use_input_pairs:
            self.experiment.evaluate(metric_name, eval_fn, self.input_pairs_dict)
        else:
            self.experiment.evaluate(metric_name, eval_fn)

    def gather_feedback(self) -> None:
        self.experiment.gather_feedback(self.input_pairs_dict, self.PIVOT_COLUMNS)

    def visualize(self, pivot: bool = False) -> None:
        r"""
        Displays a visualization of the experiment results.
        """
        if pivot:
            self.experiment.visualize(self.input_pairs_dict, self.PIVOT_COLUMNS)
        else:
            self.experiment.visualize()

    def rank(self, metric_name: str, is_average: bool = False) -> Dict[str, float]:
        r"""
        Scores and ranks the experiment inputs using the pivot columns,
        e.g. prompt templates or system prompts.
        """
        return self.experiment.rank(
            self.input_pairs_dict, self.PIVOT_COLUMNS, metric_name, is_average
        )



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List, Tuple
import csv
import logging

from prompttools.prompttest.threshold_type import ThresholdType
from prompttools.prompttest.error.failure import log_failure
from prompttools.harness import SystemPromptExperimentationHarness
from prompttools.prompttest.runner.runner import PromptTestRunner
from prompttools.prompttest.error.failure import PromptTestSetupException


class SystemPromptTestRunner(PromptTestRunner):
    r"""
    A prompt test runner for tests based on system prompts.
    """

    def __init__(self):
        self.system_prompts = {}
        self.human_messages = {}
        super().__init__()

    def read(
        self, system_prompt_file: str, human_messages_file: str
    ) -> Tuple[str, List[List[str]]]:
        r"""
        Reads data from files and parses it into a system prompt and human messages.
        """
        if (
            system_prompt_file in self.system_prompts
            and human_messages_file in self.human_messages
        ):
            return (
                self.system_prompts[system_prompt_file],
                self.human_messages[human_messages_file],
            )
        with open(system_prompt_file, "r") as f:
            system_prompt = f.read()
        with open(human_messages_file) as f:
            human_messages = list(csv.reader(f))
        self.system_prompts[system_prompt_file] = system_prompt
        self.human_messages[human_messages_file] = human_messages
        return system_prompt, human_messages

    @staticmethod
    def _get_harness(
        experiment_classname,
        model_name: str,
        system_prompt: str,
        human_messages: List[str],
        model_args: Dict[str, object],
    ) -> SystemPromptExperimentationHarness:
        return SystemPromptExperimentationHarness(
            experiment_classname, model_name, [system_prompt], human_messages, model_arguments=model_args
        )


system_prompt_test_runner = SystemPromptTestRunner()


def run_system_prompt_test(
    experiment_classname,
    model_name: str,
    metric_name: str,
    eval_fn: Callable,
    threshold: float,
    threshold_type: ThresholdType,
    is_average: bool,
    system_prompt: str,
    human_messages: List[str],
    use_input_pairs: bool,
    model_args: Dict[str, object],
) -> int:
    r"""
    Runs the prompt test.
    """
    key = system_prompt_test_runner.run(
        experiment_classname, model_name, system_prompt, human_messages, model_args
    )
    system_prompt_test_runner.evaluate(key, metric_name, eval_fn, use_input_pairs)
    scored_template = system_prompt_test_runner.rank(key, metric_name, is_average)
    if not scored_template:
        logging.error(
            "Something went wrong during testing. Make sure your API keys are set correctly."
        )
        raise PromptTestSetupException
    if (
        scored_template[system_prompt] < threshold
        and threshold_type is ThresholdType.MINIMUM
    ):
        log_failure(
            metric_name,
            threshold,
            actual=scored_template[system_prompt],
            threshold_type=threshold_type,
        )
        return 1
    if (
        scored_template[system_prompt] > threshold
        and threshold_type is ThresholdType.MAXIMUM
    ):
        log_failure(
            metric_name,
            threshold,
            actual=scored_template[system_prompt],
            threshold_type=threshold_type,
        )
        return 1
    return 0


def run_system_prompt_test_from_files(
    experiment_classname,
    model_name: str,
    metric_name: str,
    eval_fn: Callable,
    threshold: float,
    threshold_type: ThresholdType,
    is_average: bool,
    system_prompt_file: str,
    human_messages_file: str,
    use_input_pairs: bool,
    model_args: Dict[str, object],
) -> int:
    r"""
    Reads data in from files and runs the prompt test.
    """
    system_prompt, human_messages = system_prompt_test_runner.read(
        system_prompt_file, human_messages_file
    )
    return run_system_prompt_test(
        experiment_classname,
        model_name,
        metric_name,
        eval_fn,
        threshold,
        threshold_type,
        is_average,
        system_prompt,
        human_messages,  # TODO: The type of this may be incorrect
        use_input_pairs,
        model_args,
    )



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional
from .harness import ExperimentationHarness
from prompttools.experiment import OpenAIChatExperiment


class ChatModelComparisonHarness(ExperimentationHarness):
    r"""
    An experimentation harness used for comparing chat models.
    Multi-model version of ``ChatHistoryExperimentationHarness``.

    Args:
        model_names (List[str]): The names of the models that you would like to compare
        chat_histories (List[List[Dict[str, str]]]): A list of chat histories that will be fed into the models.
        runs (int): Number of runs to execute. Defaults to ``1``.
        model_arguments (Optional[Dict[str, object]], optional): Additional arguments for the model.
            Defaults to ``None``.
    """

    PIVOT_COLUMNS = ["model", "messages"]

    def __init__(
        self,
        model_names: List[str],
        chat_histories: List[List[Dict[str, str]]],
        runs: int = 1,
        model_arguments: Optional[Dict[str, object]] = None,
    ):
        self.experiment_classname = OpenAIChatExperiment
        self.model_names = model_names
        self.chat_histories = chat_histories
        self.runs = runs
        self.model_arguments = {} if model_arguments is None else model_arguments
        super().__init__()

    def prepare(self) -> None:
        """
        Initializes and prepares the experiment.
        """
        self.experiment = self.experiment_classname(
            self.model_names,
            self.chat_histories,
            **self._prepare_arguments(self.model_arguments),
        )
        super().prepare()

    def run(self):
        if not self.experiment:
            self.prepare()
        super().run()

    def compare(self):
        self.experiment.compare(self.model_names[0], self.PIVOT_COLUMNS)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict, List, Optional
import openai


from prompttools.mock.mock import mock_chat_completion_fn
from .experiment import Experiment


class OpenAIChatExperiment(Experiment):
    r"""
    This class defines an experiment for OpenAI's chat completion API.
    It accepts lists for each argument passed into OpenAI's API, then creates
    a cartesian product of those arguments, and gets results for each.
    """

    PARAMETER_NAMES = ["model", "prompt"]

    def __init__(
        self,
        model: List[str],
        prompt: List[str],
        **kwargs: Dict[str, object],
    ):
        self.completion_fn = openai.ChatCompletion.create
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_chat_completion_fn
        self.all_args = []
        self.all_args.append(model)
        self.all_args.append(prompt)
        for k, v in kwargs.items():
            self.PARAMETER_NAMES.append(k)
            self.all_args.append(v)
        super().__init__()

    @staticmethod
    def _extract_responses(output: Dict[str, object]) -> list[str]:
        return [choice["message"]["content"] for choice in output["choices"]]

    def _is_chat(self):
        return True


# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List, Optional, Tuple
from collections import defaultdict
import itertools
import logging
from IPython import display
from tabulate import tabulate
import pandas as pd

from prompttools.requests.request_queue import RequestQueue
from ..widgets.feedback import FeedbackWidgetProvider
from ..widgets.comparison import ComparisonWidgetProvider
from ..widgets.utility import is_interactive
from .error import PromptExperimentException

pd.set_option("display.max_colwidth", 0)


class Experiment:
    r"""
    Base class for experiment. This should not be used directly, please use the subclasses instead.
    """

    completion_fn: Callable
    PARAMETER_NAMES: tuple
    all_args: list

    def __init__(self):
        self.queue = RequestQueue()
        self.argument_combos = []
        self.results = []
        self.scores = defaultdict(list)
        self.feedback_widget_provider = FeedbackWidgetProvider(
            self.completion_fn, self._aggregate_metric, self._get_human_eval_listener
        )
        self.comparison_widget_provider = ComparisonWidgetProvider(
            self.completion_fn,
            self._aggregate_comparison,
            self._get_comparison_listener,
        )

    def _is_chat(self):
        return False
    
    def _get_human_eval_listener(self, i: int) -> Callable:
        def listener(change):
            self.scores["feedback"][i] = change["new"]

        return listener

    def _get_comparison_listener(self, index: int) -> Callable:
        # TODO: Map this index to the original index of the primary model
        # so we can submit feedback for the row
        def listener(change):
            new_index = self.comparison_index_translation(index)
            self.scores["comparison"][new_index] = change["new"]

        return listener

    def _aggregate_comparison(
        self,
        table: pd.DataFrame,
        agg_column: int = 0,
        is_average: bool = False,
    ) -> Dict[str, int]:
        # TODO: This could be a group by
        prompt_scores = defaultdict(int)
        prompt_counts = defaultdict(int)
        for index, row in enumerate(table.iterrows()):
            key = str(row[agg_column])
            new_index = self.comparison_index_translation(index)
            prompt_scores[key] += self.scores["comparison"][new_index]
            prompt_counts[key] += 1
        if is_average:
            for k, v in prompt_scores.items():
                prompt_scores[k] = v / prompt_counts[k]
        sorted_scores = dict(
            sorted(prompt_scores.items(), key=lambda item: item[1], reverse=True)
        )
        return sorted_scores

    def _aggregate_metric(
        self,
        table: pd.DataFrame,
        metric_name: str,
        agg_column: str,
        is_average: bool = False,
    ) -> Dict[str, int]:
        # TODO: This could be a group by

        prompt_scores = defaultdict(int)
        prompt_counts = defaultdict(int)
        for index, row in table.iterrows():
            key = str(row[agg_column])
            prompt_scores[key] += self.scores[metric_name][index]
            prompt_counts[key] += 1
        if is_average:
            for k, v in prompt_scores.items():
                prompt_scores[k] = v / prompt_counts[k]
        sorted_scores = dict(
            sorted(prompt_scores.items(), key=lambda item: item[1], reverse=True)
        )
        return sorted_scores

    def _create_args_dict(
        self, args: Dict[str, object], tagname: str, input_pairs: Dict[str, object]
    ) -> Dict[str, object]:
        args = {self.PARAMETER_NAMES[i]: arg for i, arg in enumerate(args)}
        return {name: arg for name, arg in args.items() if arg and arg != float("inf")}

    def prepare(self) -> None:
        r"""
        Creates argument combinations by taking the cartesian product of all inputs.
        """
        self.argument_combos = []
        for combo in itertools.product(*self.all_args):
            self.argument_combos.append(combo)

    def run(
        self,
        tagname: Optional[str] = "",
        input_pairs: Optional[Dict[str, Tuple[str, Dict[str, str]]]] = None,
        runs: int = 1,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.
        For each combination, it will execute `runs` times, default to 1.
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        for combo in self.argument_combos:
            for _ in range(runs):
                self.queue.enqueue(
                    self.completion_fn,
                    self._create_args_dict(combo, tagname, input_pairs),
                )
        self.results = self.queue.results()
        self.scores["latency"] = self.queue.latencies()
        if len(self.results) == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException

    def evaluate(
        self,
        metric_name: str,
        eval_fn: Callable,
        input_pairs: Optional[Dict[str, Tuple[str, Dict[str, str]]]] = None,
    ) -> None:
        """
        Using the given evaluation function, all input/response pairs are evaluated.
        """
        if not self.results:
            logging.info("Running first...")
            self.run()
        if metric_name in self.scores:
            logging.warning(metric_name + " is already present, skipping.")
            return
        for i, result in enumerate(self.results):
            # Pass the messages and results into the eval function
            score = eval_fn(
                input_pairs[self.argument_combos[i][1]]
                if input_pairs
                else self.argument_combos[i][1],
                result,
                {
                    name: self.scores[name][i]
                    for name in self.scores.keys()
                    if name is not metric_name
                },
            )
            self.scores[metric_name].append(score)

    def get_table(
        self, pivot_data: Dict[str, object], pivot_columns: List[str], pivot: bool
    ) -> pd.DataFrame:
        """
        This method creates a table of the experiment data. It can also be used
        to create a pivot table, or a table for gathering human feedback.
        """
        data = {
            "messages" if self._is_chat() else "prompt": [str(combo[1]) for combo in self.argument_combos],
            "response(s)": [self._extract_responses(result) for result in self.results],
            "latency": self.scores["latency"],
        }
        # Add scores for each eval fn, including feedback
        for metric_name, evals in self.scores.items():
            if metric_name != "comparison":
                data[metric_name] = evals
        # Add other args as cols if there was more than 1 input
        for i, args in enumerate(self.all_args):
            if len(args) > 1:
                data[self.PARAMETER_NAMES[i]] = [
                    combo[i] for combo in self.argument_combos
                ]
        if pivot_data:
            data[pivot_columns[0]] = [
                str(pivot_data[str(combo[1])][0]) for combo in self.argument_combos
            ]
            data[pivot_columns[1]] = [
                str(pivot_data[str(combo[1])][1]) for combo in self.argument_combos
            ]
        df = pd.DataFrame(data)
        if pivot:
            df = pd.pivot_table(
                df,
                values="response(s)",
                index=[pivot_columns[1]],
                columns=[pivot_columns[0]],
                aggfunc=lambda x: x.iloc[0],
            )
        return df

    def gather_feedback(
        self, pivot_data: Dict[str, object], pivot_columns: List[str]
    ) -> None:
        """
        This method creates a table to gather human feedback from a notebook interface.
        """
        if not self.results:
            logging.info("Running first...")
            self.run()
        if not is_interactive():
            logging.warning("This method only works in notebooks.")
            return
        self.scores["feedback"] = [1] * len(self.results)
        table = self.get_table(pivot_data, pivot_columns, pivot=False)
        self.feedback_widget_provider.set_pivot_columns(pivot_columns)
        items = self.feedback_widget_provider.get_header_widgets()
        for row in table.iterrows():
            items += self.feedback_widget_provider.get_row_widgets(*row)
        items += self.feedback_widget_provider.get_footer_widgets(table)
        self.feedback_widget_provider.display(items)

    def compare(self, primary_model: str, pivot_columns: List[str]) -> None:
        """
        This method creates a table to gather human feedback from a notebook interface.
        """
        if not self.results:
            logging.info("Running first...")
            self.run()
        if not is_interactive():
            logging.warning("This method only works in notebooks.")
            return
        table = self.get_table(pivot_data={}, pivot_columns=pivot_columns, pivot=True)
        self.scores["comparison"] = [1] * len(table)
        self.comparison_index_translation = lambda i: i * len(table.columns)
        self.comparison_widget_provider.set_models(table.columns)
        items = self.comparison_widget_provider.get_header_widgets()
        for index, row in enumerate(table.iterrows()):
            items += self.comparison_widget_provider.get_row_widgets(index, row[1])
        items += self.comparison_widget_provider.get_footer_widgets(table)
        self.comparison_widget_provider.display(items)

    def visualize(
        self,
        pivot_data: Optional[Dict[str, object]] = None,
        pivot_columns: Optional[List[str]] = None,
    ) -> None:
        """
        Creates and shows a table using the results produced.
        """
        if not self.results:
            logging.info("Running first...")
            self.run()
        table = self.get_table(
            pivot_data, pivot_columns, pivot=pivot_columns is not None
        )
        if is_interactive():
            display.display(table)
        else:
            logging.getLogger().setLevel(logging.INFO)
            logging.info(tabulate(table, headers="keys", tablefmt="psql"))

    def rank(
        self,
        pivot_data: Dict[str, object],
        pivot_columns: List[str],
        metric_name: str,
        is_average: bool,
    ) -> Dict[str, int]:
        """
        Using pivot data, groups the data by the first pivot column to
        get scores, and sorts descending. For example, using pivot data of
        (prompt_template, user_input), a metric of latency, and is_average=True,
        we rank prompt templates by their average latency in the test set.
        """
        if metric_name not in self.scores:
            logging.warning(
                "Can't find " + metric_name + " in scores. Did you run `evaluate`?"
            )
            return
        table = self.get_table(pivot_data, pivot_columns, pivot=False)
        sorted_scores = self._aggregate_metric(
            table, metric_name, pivot_columns[0], is_average
        )
        return sorted_scores

    @staticmethod
    def _extract_responses(output: Dict[str, object]) -> list[str]:
        raise NotImplementedError(
            "This should be implemented by a subclass of `Experiment`."
        )



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict, List, Optional
from .harness import ExperimentationHarness


class SystemPromptExperimentationHarness(ExperimentationHarness):
    r"""
    An experimentation harness used to test various system prompts.

    Args:
        model_name (str): The name of the model.
        system_prompts (List[str]): A list of system prompts
        human_messages (List[str]): A list of
        model_arguments (Optional[Dict[str, object]], optional): Additional arguments for the model.
            Defaults to ``None``.
    """

    PIVOT_COLUMNS = ["system_prompt", "user_input"]

    def __init__(
        self,
        experiment_classname,
        model_name: str,
        system_prompts: List[str],
        human_messages: List[str],
        model_arguments: Optional[Dict[str, object]] = None,
    ):
        self.experiment_classname = experiment_classname,
        self.model_name = model_name
        self.system_prompts = system_prompts
        self.human_messages = human_messages
        self.model_arguments = {} if model_arguments is None else model_arguments
        super().__init__()

    @staticmethod
    def _create_system_prompt(content: str) -> Dict[str, str]:
        return {"role": "system", "content": content}

    @staticmethod
    def _create_human_message(content: str) -> Dict[str, str]:
        return {"role": "user", "content": content}

    def prepare(self) -> None:
        r"""
        Creates messages to use for the experiment, and then initializes and prepares the experiment.
        """
        self.input_pairs_dict = {}
        messages_to_try = []
        for system_prompt in self.system_prompts:
            for message in self.human_messages:
                history = [
                    self._create_system_prompt(system_prompt),
                    self._create_human_message(message),
                ]
                messages_to_try.append(history)
                self.input_pairs_dict[str(history)] = (system_prompt, message)
        self.experiment = self.experiment_classname(
            [self.model_name],
            messages_to_try,
            **self._prepare_arguments(self.model_arguments),
        )
        super().prepare()

    def run(self):
        if not self.experiment:
            self.prepare()
        super().run()



from unittest import TestCase

from prompttools.harness import (
    ChatHistoryExperimentationHarness,
    ChatModelComparisonHarness,
    PromptTemplateExperimentationHarness,
    SystemPromptExperimentationHarness,
)


class TestHarness(TestCase):
    # TODO: Currently, it only ensures importing is correct.
    #       Add unit tests to verify initialization.
    def test_chat_history_exp_harness(self):
        pass

    def test_chat_model_exp_harness(self):
        pass

    def test_prrmpt_template_exp_harness(self):
        pass

    def test_system_prompt_exp_harness(self):
        pass



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, List, Optional, Tuple

import logging

from llama_cpp import Llama
from time import perf_counter

from .experiment import Experiment
from .error import PromptExperimentException


class LlamaCppExperiment(Experiment):
    r"""
    Used to experiment across parameters for a local model, supported by LlamaCpp and GGML.
    """

    PARAMETER_NAMES = (
        "model_path",
        "prompt",
        "lora_path",
        "lora_base",
        "n_ctx",
        "n_parts",
        "seed",
        "f16_kv",
        "logits_all",
        "vocab_only",
        "use_mlock",
        "n_threads",
        "n_batch",
        "use_mmap",
        "last_n_tokens_size",
        "verbose",
        "suffix",
        "max_tokens",
        "temperature",
        "top_p",
        "logprobs",
        "echo",
        "stop",
        "repeat_penalty",
        "top_k",
    )

    MODEL_PARAMETERS = PARAMETER_NAMES[2:15]
    CALL_PARAMETERS = ("prompt",) + PARAMETER_NAMES[16:]

    DEFAULT = {
        "lora_path": [None],
        "lora_base": [None],
        "n_ctx": [512],
        "n_parts": [-1],
        "seed": [1337],
        "f16_kv": [True],
        "logits_all": [False],
        "vocab_only": [False],
        "use_mlock": [False],
        "n_threads": [None],
        "n_batch": [512],
        "use_mmap": [True],
        "last_n_tokens_size": [64],
        "verbose": [True],
        "suffix": [None],
        "max_tokens": [128],
        "temperature": [0.8],
        "top_p": [0.95],
        "logprobs": [None],
        "echo": [False],
        "stop": [None],
        "repeat_penalty": [1.1],
        "top_k": [40],
    }

    def __init__(
        self, model_path: List[str], prompt: List[str], **kwargs: Dict[str, object]
    ):
        self.completion_fn = self.llama_completion_fn
        self.all_args = []
        self.all_args.append(model_path)
        self.all_args.append(prompt)
        for param in self.PARAMETER_NAMES[2:]:
            if param in kwargs:
                self.all_args.append(kwargs[param])
            elif param in self.DEFAULT:
                self.all_args.append(self.DEFAULT[param])
            elif param == "prompt":
                self.all_args.append(prompt)
        super().__init__()

    def llama_completion_fn(
        self,
        **params: Dict[str, Any],
    ):
        r"""
        Local model helper function to make request
        """
        model_params = {k: v for k, v in params.items() if k in self.MODEL_PARAMETERS}
        call_params = {k: v for k, v in params.items() if k in self.CALL_PARAMETERS}
        client = Llama(model_path=params["model_path"], **model_params)
        response = client(**call_params)
        logging.info(response)
        return response

    def run(
        self,
        tagname: Optional[str] = "",
        input_pairs: Optional[Dict[str, Tuple[str, Dict[str, str]]]] = None,
        runs: int = 1,
    ) -> None:
        r"""
        Create tuples of input and output for every possible combination of arguments.
        For each combination, it will execute `runs` times, default to 1.
        For local models we need to run this in a single thread.
        """
        if not self.argument_combos:
            logging.info("Preparing first...")
            self.prepare()
        for combo in self.argument_combos:
            for _ in range(runs):
                start = perf_counter()
                res = self.completion_fn(
                    **self._create_args_dict(combo, tagname, input_pairs)
                )
                self.results.append(res)
                self.scores["latency"] = perf_counter() - start
        if len(self.results) == 0:
            logging.error("No results. Something went wrong.")
            raise PromptExperimentException

    @staticmethod
    def _extract_responses(output: Dict[str, object]) -> list[str]:
        return [choice["text"] for choice in output["choices"]]



from typing import List
import json
import re

KEY_EXTRACTION_REGEX = r'"([^"]+?)"\s*:'

def validate(text: str):
    r"""
    Validates that the generated text is JSON.

    Args:
        text (str): The generated text, which should be valid JSON.
    """
    try:
        json.loads(text)
    except ValueError:
        return 0.0
    return 1.0

def validate_keys(text: str, valid_keys: List[str]):
    r"""
    Guarantees that all keys in the generated JSON are valid.

    Args:
        text (str): The generated text, which should be valid JSON.
        valid_keys (List[str]): A list of valid keys which may appear in the JSON.
    """
    match = re.search(text, KEY_EXTRACTION_REGEX)
    for group in match.groups():
        if group not in valid_keys:
            return 0.0
    return 1.0
    


import os
import openai
from .error import PromptToolsUtilityError
import similarity

def compute(prompt: str, model: str = 'gpt-4') -> str:
    r"""
    Computes the expected result of a given prompt by using a high
    quality LLM, like GPT-4.

    Args:
        prompt (str): The input prompt.
        model (str): The OpenAI chat model to use for generating an expected response.
            Defaults to GPT-4.
    """
    if not os.environ['OPENAI_API_KEY']:
        raise PromptToolsUtilityError
    response = openai.ChatCompletion.create(model=model,
                                            prompt=prompt)
    return response['choices'][0]['message']['content']

def compute_similarity(prompt: str, response: str, model: str = 'gpt-4') -> str:
    r"""
    Computes the similarity of a given response to the expected result 
    of a given prompt by using a high quality LLM, like GPT-4.

    Args:
        prompt (str): The input prompt.
        response (str): The model response.
        model (str): The OpenAI chat model to use for generating an expected response.
            Defaults to GPT-4.
    """
    expected_response = compute(prompt, model)
    return similarity.compute(response, expected_response)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from enum import Enum


class ThresholdType(Enum):
    r"""
    Defines the types of thresholds a user can specify for their test case.
    """

    MINIMUM = 1
    MAXIMUM = 2



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable, Dict, List, Tuple
import csv
import logging

from prompttools.prompttest.threshold_type import ThresholdType
from prompttools.prompttest.error.failure import log_failure
from prompttools.harness import PromptTemplateExperimentationHarness
from prompttools.prompttest.runner.runner import PromptTestRunner
from prompttools.prompttest.error.failure import PromptTestSetupException


class PromptTemplateTestRunner(PromptTestRunner):
    r"""
    A prompt test runner for tests based on prompt templates.
    """

    def __init__(self):
        self.prompt_templates = {}
        self.user_inputs = {}
        super().__init__()

    def read(
        self, prompt_template_file: str, user_input_file: str
    ) -> Tuple[str, List[Dict[str, str]]]:
        """
        Reads data from files and parses it into a prompt template and user input.
        """
        if (
            prompt_template_file in self.prompt_templates
            and user_input_file in self.user_inputs
        ):
            return (
                self.prompt_templates[prompt_template_file],
                self.user_inputs[user_input_file],
            )
        prompt_template = ""
        with open(prompt_template_file, "r") as f:
            prompt_template = f.read()
        user_inputs = [{"": ""}]
        with open(user_input_file) as f:
            user_inputs = list(csv.DictReader(f))
        self.prompt_templates[prompt_template_file] = prompt_template
        self.user_inputs[user_input_file] = user_inputs
        return prompt_template, user_inputs

    @staticmethod
    def _get_harness(
        experiment_classname,
        model_name: str,
        prompt_template: str,
        user_inputs: List[Dict[str, str]],
        model_args: Dict[str, object],
    ) -> PromptTemplateExperimentationHarness:
        return PromptTemplateExperimentationHarness(
            experiment_classname, model_name, [prompt_template], user_inputs, model_arguments=model_args
        )


prompt_template_test_runner = PromptTemplateTestRunner()


def run_prompt_template_test(
    experiment_classname,
    model_name: str,
    metric_name: str,
    eval_fn: Callable,
    threshold: float,
    threshold_type: ThresholdType,
    is_average: bool,
    prompt_template: str,
    user_inputs: List[Dict[str, str]],
    use_input_pairs: bool,
    model_args: Dict[str, object],
) -> int:
    """
    Runs the prompt test.
    """
    key = prompt_template_test_runner.run(
        experiment_classname, model_name, prompt_template, user_inputs, model_args
    )
    prompt_template_test_runner.evaluate(key, metric_name, eval_fn, use_input_pairs)
    scored_template = prompt_template_test_runner.rank(key, metric_name, is_average)
    if not scored_template:
        logging.error(
            "Something went wrong during testing. Make sure your API keys are set correctly."
        )
        raise PromptTestSetupException
    if (
        scored_template[prompt_template] < threshold
        and threshold_type is ThresholdType.MINIMUM
    ):
        log_failure(
            metric_name,
            threshold,
            actual=scored_template[prompt_template],
            threshold_type=threshold_type,
        )
        return 1
    if (
        scored_template[prompt_template] > threshold
        and threshold_type is ThresholdType.MAXIMUM
    ):
        log_failure(
            metric_name,
            threshold,
            actual=scored_template[prompt_template],
            threshold_type=threshold_type,
        )
        return 1
    return 0


def run_prompt_template_test_from_files(
    experiment_classname,
    model_name: str,
    metric_name: str,
    eval_fn: Callable,
    threshold: float,
    threshold_type: ThresholdType,
    is_average: bool,
    prompt_template_file: str,
    user_input_file: str,
    use_input_pairs: bool,
    model_args: Dict[str, object],
) -> int:
    """
    Reads data in from files and runs the prompt test.
    """
    prompt_template, user_inputs = prompt_template_test_runner.read(
        prompt_template_file, user_input_file
    )
    return run_prompt_template_test(
        experiment_classname,
        model_name,
        metric_name,
        eval_fn,
        threshold,
        threshold_type,
        is_average,
        prompt_template,
        user_inputs,
        use_input_pairs,
        model_args,
    )



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.


from .experiments.experiment import Experiment
from .experiments.openai_chat_experiment import OpenAIChatExperiment
from .experiments.openai_completion_experiment import OpenAICompletionExperiment
from .experiments.huggingface_hub_experiment import HuggingFaceHubExperiment
from .experiments.llama_cpp_experiment import LlamaCppExperiment


__all__ = [
    "Experiment", 
    "LlamaCppExperiment", 
    "HuggingFaceHubExperiment", 
    "OpenAIChatExperiment", 
    "OpenAICompletionExperiment"
]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Dict, List
from huggingface_hub.inference_api import InferenceApi

from prompttools.mock.mock import mock_hf_completion_fn

from .experiment import Experiment

VALID_TASKS = ("text2text-generation", "text-generation", "summarization")


class HuggingFaceHubExperiment(Experiment):
    r"""
    Experiment for Hugging Face Hub's API. 
    It accepts lists for each argument passed into Hugging Face Hub's API, 
    then creates a cartesian product of those arguments, and gets results for each.
    """

    PARAMETER_NAMES = ["repo_id", "prompt", "task"]

    def __init__(
        self,
        repo_id: List[str],
        prompt: List[str],
        task: List[str] = ["text-generation"],
        **kwargs: Dict[str, object],
    ):
        self.completion_fn = self.hf_completion_fn
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_hf_completion_fn
        self.all_args = []
        self.all_args.append(repo_id)
        self.all_args.append(prompt)
        self.all_args.append(task)
        for k, v in kwargs.items():
            self.PARAMETER_NAMES.append(k)
            self.all_args.append(v)
        super().__init__()

    def hf_completion_fn(
        self,
        **params: Dict[str, Any],
    ):
        r"""
        Hugging Face Hub helper function to make request
        """
        client = InferenceApi(
            repo_id=params["repo_id"],
            token=os.environ.get("HUGGINGFACEHUB_API_TOKEN"),
            task=params["task"],
        )
        model_kwargs = {
            k: params[k] for k in params if k not in ["repo_id", "prompt", "task"]
        }
        response = client(inputs=params["prompt"], params=model_kwargs)
        return response

    @staticmethod
    def _extract_responses(output: Dict[str, object]) -> list[str]:
        return output["generated_text"]



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

r"""
Use a list to optionally hold a reference to the embedding model and client,
allowing for lazy initialization.
"""
EMBEDDING_MODEL = []  #
CHROMA_CLIENT = []


def _get_embedding_model():
    if len(EMBEDDING_MODEL) == 0:
        from sentence_transformers import SentenceTransformer

        EMBEDDING_MODEL.append(
            SentenceTransformer("sentence-transformers/all-MiniLM-L6-v2")
        )
    return EMBEDDING_MODEL[0]


def _get_chroma_client():
    if len(CHROMA_CLIENT) == 0:
        import chromadb

        CHROMA_CLIENT.append(chromadb.Client())
    return CHROMA_CLIENT[0]


def _from_huggingface(doc1, doc2):
    model = _get_embedding_model()
    embedding_1 = model.encode(doc1, convert_to_tensor=True)
    embedding_2 = model.encode(doc2, convert_to_tensor=True)
    from sentence_transformers.util import pytorch_cos_sim

    return pytorch_cos_sim(embedding_1, embedding_2).item()


def _from_chroma(doc1, doc2):
    chroma_client = _get_chroma_client()
    collection = chroma_client.create_collection(name="test_collection")
    collection.add(documents=[doc1], ids=["id1"])
    query_results = collection.query(query_texts=doc2, n_results=1)
    chroma_client.delete_collection("test_collection")
    return query_results["distances"][0][0] / 2


def compute(doc1, doc2, use_chroma=True):
    r"""
    Computes the semantic similarity between two documents, using either ChromaDB
    or HuggingFace sentence_transformers.

    Args:
        doc1 (str): The first document.
        doc2 (str): The second document.
        use_chroma (bool): Indicates whether or not to use Chroma. 
            If false, uses HuggingFace sentence_transformers.
    """
    if use_chroma:
        return _from_chroma(doc1, doc2)
    else:
        return _from_huggingface(doc1, doc2)



# Copyright (c) Hegel AI, Inc.
# All rights reserved.
#
# This source code's license can be found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Dict, List, Optional
import openai

from prompttools.mock.mock import mock_completion_fn
from .experiment import Experiment


class OpenAICompletionExperiment(Experiment):
    r"""
    This class defines an experiment for OpenAI's chat completion API.
    It accepts lists for each argument passed into OpenAI's API, then creates
    a cartesian product of those arguments, and gets results for each.
    """

    PARAMETER_NAMES = (
        "model",
        "prompt",
        "suffix",
        "max_tokens",
        "temperature",
        "top_p",
        "n",
        "stream",
        "logprobs",
        "echo",
        "stop",
        "presence_penalty",
        "frequency_penalty",
        "best_of",
        "logit_bias",
    )

    def __init__(
        self,
        model: List[str],
        prompt: List[str],
        suffix: Optional[List[str]] = [None],
        max_tokens: Optional[List[int]] = [float("inf")],
        temperature: Optional[List[float]] = [1.0],
        top_p: Optional[List[float]] = [1.0],
        n: Optional[List[int]] = [1],
        stream: Optional[List[bool]] = [False],
        logprobs: Optional[List[int]] = [None],
        echo: Optional[List[bool]] = [False],
        stop: Optional[List[List[str]]] = [None],
        presence_penalty: Optional[List[float]] = [0],
        frequency_penalty: Optional[List[float]] = [0],
        best_of: Optional[List[int]] = [1],
        logit_bias: Optional[Dict] = [None],
    ):
        self.completion_fn = openai.Completion.create
        if os.getenv("DEBUG", default=False):
            self.completion_fn = mock_completion_fn
        self.all_args = []
        self.all_args.append(model)
        self.all_args.append(prompt)
        self.all_args.append(suffix)
        self.all_args.append(max_tokens)
        self.all_args.append(temperature)
        self.all_args.append(top_p)
        self.all_args.append(n)
        self.all_args.append(stream)
        self.all_args.append(logprobs)
        self.all_args.append(echo)
        self.all_args.append(stop)
        self.all_args.append(presence_penalty)
        self.all_args.append(frequency_penalty)
        self.all_args.append(best_of)
        self.all_args.append(logit_bias)
        super().__init__()

    @staticmethod
    def _extract_responses(output: Dict[str, object]) -> list[str]:
        return [choice["text"] for choice in output["choices"]]
